#include <ctype.h>
#include <curses.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "flitview.h"
#include "list.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

//
// input format
//
// [SimMc ver]\n
// [noc_col] [noc_row]\n
// [log_t][log_t]...[log_t]EOF
//

/***************************************************************************/
flitview::flitview() {
    fname = NULL;
    clk = phase = 0;
    current = 0;
    win = NULL;
    log = NULL;
}

/***************************************************************************/
flitview::~flitview() {
    delete [] *win;
    delete [] win;
    delete log;
}

/***************************************************************************/
#define erase_buffer()                          \
    (move(term_lines, 0),                       \
     ({for(int i=0; i<term_lines; i++)          \
             addch(' ');}))

/***************************************************************************/
int flitview::compare(log_t *p, uint clk, uint phase) {
    if(p->clk == clk)
        return p->phase - phase;
    return p->clk - clk;
}    

/***************************************************************************/
void flitview::initscreen() {
    enum{
        NONE,
        SCR_MAIN,
        SCR_SHADOW,
        SCR_ROUTER,
        SCR_ROUTER_ALT
    };
    
    initscr();
    getmaxyx(stdscr, term_lines, term_columns);

    const uint router_height = 8, router_width = 14;
    const uint router_base_y = term_lines/2 - router_height*noc_row/2;
    const uint router_base_x = term_columns/2 - router_width*noc_col/2;
    const uint margin_v = 1, margin_h = 1;
    const uint shadow_y = 1, shadow_x = 1;
    
    curs_set(0);
    noecho();
    
    // create windows
    win = new WINDOW** [noc_row];
    {
        WINDOW **_win = new WINDOW* [noc_col*noc_row];
        for(int i = 0; i < noc_row; i++) 
            win[i] = _win + i*noc_col;
    }

    for(uint i = 0; i < noc_row; i++) {
        for(uint j = 0; j < noc_col; j++) {
            win[i][j] = subwin(stdscr, router_height, router_width,
                               router_base_y + i*router_height, router_base_x + j*router_width);
            box(win[i][j], ACS_VLINE, ACS_HLINE);
            wmove(win[i][j], 0, 2);
            wprintw(win[i][j], "%02d-%02d", j, i);
        }
    }
    
    start_color();

    WINDOW *board, *shadow;
    board = subwin(stdscr, router_height*noc_row + 2*margin_v, router_width*noc_col + 2*margin_h,
                   router_base_y - margin_v, router_base_x - margin_h);
    shadow = subwin(stdscr, router_height*noc_row + 2*margin_v, router_width*noc_col + 2*margin_h,
                    router_base_y - margin_v + shadow_y, router_base_x - margin_h + shadow_x);
    init_pair(SCR_MAIN, COLOR_YELLOW, COLOR_BLUE);
    init_pair(SCR_ROUTER, COLOR_BLACK, COLOR_WHITE);
    init_pair(SCR_SHADOW, COLOR_BLACK, COLOR_BLACK);
    init_pair(SCR_ROUTER_ALT, COLOR_RED, COLOR_WHITE);
    bkgd(COLOR_PAIR(SCR_MAIN));
    attrset(A_BOLD);
    wbkgd(shadow, COLOR_PAIR(SCR_SHADOW));
    wbkgd(board, COLOR_PAIR(SCR_ROUTER));
    box(board, ACS_VLINE, ACS_HLINE);

    for(uint i = 0; i < noc_col*noc_row; i++)
        wbkgd((*win)[i], COLOR_PAIR(SCR_ROUTER));

    printw("%s %s", simver, FLITVIEW_VER);
}

/***************************************************************************/
void flitview::clear() {
    for(uint j = 0; j < noc_row; j++)
        for(uint k = 0; k < noc_col; k++)
            for(uint m = 0; m < NUM_DIRECTION; m++)
                for(uint n = 0; n < 3; n++)
                    wmove(win[j][k], coordinates[m][n][0], coordinates[m][n][1]),
                        wprintw(win[j][k], "__");
}

/***************************************************************************/
void flitview::update() {
    // refresh sub-windows
    for(uint j = 0; j < noc_row; j++)
        for(uint k = 0; k < noc_col; k++)
            wrefresh(win[j][k]);
    
    // refresh main window
    refresh();
}

/***************************************************************************/
void flitview::draw() {
    mvprintw(1, 0, "clk:%06u phase:%s", clk, phasetab[phase]);
    log_search(&current, clk, phase);

    log_t *p;

    // forward
    p = log + current;
    
//    while(compare(p, clk, phase) == 0 && p < log + logsize) {
    while(p < log + logsize && compare(p, clk, phase) == 0) {
        if(p->site > 2) {
            printf("site %u\n", p->site);
            exit(123);
        }
        const uint *c = coordinates[p->direction][p->site];
        WINDOW *w = win[p->y][p->x];
        wmove(w, c[0], c[1]);
        wprintw(w, "%02x", p->data);
        p++;
    }

    // backward
    p = log + current - 1;
//    while(compare(p, clk, phase) == 0 && p > log) {
    while(p > log && compare(p, clk, phase) == 0) {
        const uint *c = coordinates[p->direction][p->site];
        WINDOW *w = win[p->y][p->x];
        wmove(w, c[0], c[1]);
        wprintw(w, "%02x", p->data);
        p--;
    }
}

/***************************************************************************/
void flitview::phase_jump(int num) {
    int nphase = clk*3 + phase;
    clk = (nphase + num)/3;
    phase = (nphase + num)%3;
}

/***************************************************************************/
void flitview::goto_cycle(uint clk) {
    phase = 0;
    if(clk > maxclk)
        this->clk = maxclk;
    else if(clk < 0)
        this->clk = 0;
    else
        this->clk = clk;
}

/***************************************************************************/
void flitview::cursesgets() {
    curs_set(1);
    echo();
    move(term_lines - 1, 0);
    addch('/');
    move(term_lines - 1, 1);
    getnstr(str, 256);
    erase_buffer();
    curs_set(0);
    noecho();
}

/***************************************************************************/
int __compare(const void *_a, const void *_b) {
    log_t *a = (log_t*)_a;
    log_t *b = (log_t*)_b;
    
    if(a->clk == b->clk)
        return a->phase - b->phase;
    return a->clk - b->clk;
}

/***************************************************************************/
void flitview::log_search(uint *n, uint clk, uint phase) {
    log_t tmp;
    tmp.clk = clk, tmp.phase = phase;

    log_t *p = (log_t*)bsearch(&tmp, log, logsize, sizeof(log_t), __compare);
    if(p != NULL)
        *n = (uint)(p - log);
}
  
/***************************************************************************/
void flitview::search_next() {
    //added
    for(int i = 0; i < logsize; i++)
        if(compare(log + i, clk, phase) > 0) {
            current = i;
            break;
        }

    for(int i = current + 1; i < logsize; i++)
        if(log[i].clk != clk || log[i].phase != phase) {
            current = i;
            clk = log[current].clk;
            phase = log[current].phase;
            break;
        }
}

/***************************************************************************/
void flitview::search_prev() {
    //added
    for(int i = logsize - 1; i >= 0; i--)
        if(compare(log + i, clk, phase) < 0) {
            current = i;
            break;
        }

    for(int i = current - 1; i < logsize; i--)
        if(log[i].clk != clk || log[i].phase != phase) {
            current = i;
            clk = log[current].clk;
            phase = log[current].phase;
            break;
        }
}

/***************************************************************************/
void flitview::show_cui() {
    uint end = 0;

    dup2(open("/dev/tty", O_RDONLY ), STDIN_FILENO );
    initscreen();
 
    while(!end) {
        clear();
        draw();
        update();
        
        // get command
        char c;
        uint valid;
        do {
            uint num = 0;
            static uint f_num = 1;
            valid = 1;
            // get number
            if(isdigit(c = getch())) {
                num = c - '0';
                while(isdigit(c = getch()))
                    num = 10*num + c - '0';
            }
            switch(c) {
            case 'j':
                phase_jump(num == 0 ? 1 : num);
                break;
            case 'k':
                phase_jump(num == 0 ? -1 : -num);
                break;
            case 'g':
                goto_cycle(num);
                break;
            case 'f':
            case 'u':
                goto_cycle(clk + (num == 0 ? 1 : num));
                break;
            case 'b':
            case 'd':
                goto_cycle(clk - (num == 0 ? 1 : num));
                break;
            case 'G':
                goto_cycle(maxclk);
                break;
            case 'q':
                end = 1;
                break;
            case '/':
                cursesgets();
                break;
            case 'n':
                search_next();
                break;
            case 'N':
                search_prev();
               break;
            default:
                valid = 0;
                break;
            }
        } while(!valid);
    }
    endwin();
}

/***************************************************************************/
void flitview::read_log() {
    loglist_t *buf;

    if(fname == NULL)
        fp = stdin;
    else
        fp = fopen(fname, "r");

    fgets(simver, 256, fp);
    fscanf(fp, "%d %d\n", &noc_col, &noc_row);
    noc_col++, noc_row++;
    
    struct list_head loghead;
    INIT_LIST_HEAD(&loghead);
    buf = new loglist_t;
    for(logsize = 0; 
        fread(&buf->log, sizeof(log_t), 1, fp);
        logsize++) {
        list_add_tail(&buf->list, &loghead);
        buf = new loglist_t;
    }
    delete buf;

    if(list_empty(&loghead)) {
        fprintf(stderr, "## ERROR: log is empty.\n");
        exit(306);
    }

    loglist_t *end = list_entry(loghead.prev, loglist_t, list);
    maxclk = end->log.clk;
    --logsize;

    list_del(&end->list);
    delete end;

    loglist_t *p;
    
    log = new log_t [logsize];
    int i = 0;

    list_for_each_entry(p, &loghead, list)
        log[i++] = p->log;
}

/***************************************************************************/
char* checkargs(int argc, char **argv) {
    char *file = NULL;

    while(argc > 0) {
        const char *arg = *argv;
        if(*arg == '-') {
            switch(arg[1]) {
            case 'D': 
                // for SimMc old version
                break;
            case '-':
                if(!strcmp("help", arg + 2)) {
                    puts(FLITVIEW_VER);
                    fputs(USAGE, stdout);
                    exit(1);
                }
                else if(!strcmp("version", arg + 2)) {
                    puts(FLITVIEW_VER);
                    exit(1);
                }
                else {
                    printf("unknown option %s\n", arg);
                    exit(1);
                }
                break;
            }
        }
        else {
            file = *argv;
        }            
        argc--, argv++;
    }
    return file;
}

/***************************************************************************/
int main(int argc, char **argv) {
    flitview view;
    view.fname = checkargs(argc - 1, argv + 1);
    view.read_log();
    view.show_cui();

    return 0;
}
