/**********************************************************************/
/* Many-Core Architecture Research Project       Arch Lab. TOKYO TECH */
/**********************************************************************/
#include "network.h"
#include "util.h"

/**********************************************************************/
/**
 * show program usage
 */
extern void usage();

/**********************************************************************/
/**
 * assigned when signal INT occured
 */
static volatile sig_atomic_t sigint = 0;

/**********************************************************************/
/**
 * signal handler
 * \param signo signal number
 * \sa Env::initialize()
 */
static void sigint_handler(int signo) {
    sigint = 1;
}

/**********************************************************************/
/**
 * convert string including unit to integer
 * ex. "10k" -> 10000
 * \param nptr string
 * \return converted value
 */
ullint atoi_postfix(const char *nptr)
{
    char *endptr;
    ullint result = strtoll(nptr, &endptr, 10);
    if ((*endptr == 'k') || (*endptr == 'K'))
        result *= 1000;
    else if ((*endptr == 'm') || (*endptr == 'M'))
        result *= 1000000;
    else if ((*endptr == 'g') || (*endptr == 'G'))
        result *= 1000000000;
    return result;
}

/**********************************************************************/
/**
 * Env constructor
 */
Env::Env()
{
    core = NULL;
    router = NULL;
    incc32 = NULL;
    nodeinfo = NULL;
    node = NULL;
    prev_pc = NULL;
    logfile = NULL;
    rank_x = DEF_RANK_X;
    rank_y = DEF_RANK_Y;
    maxcycle = DEF_MAX_CYCLE;
    debug_mode = 0;
    add_module = 1;
    incc_activity = 0;
    log_mode = 0;
    quiet_mode = 0;
    server_mode = 0;
    cycle = 0;
    phase = 0;
}

/**********************************************************************/
/**
 * Env destructor
 */
Env::~Env()
{
    delete [] incc32;
    delete [] incc64;
    delete [] gmem;
    delete [] router;
    delete [] core;
    delete [] nodeinfo;
    delete [] node;
    delete [] prev_pc;
}

void Env::make_nodeinfo(char *filename)
{
    if(add_module)
        noc_col = rank_x + 1, noc_row = rank_y + 1;
    else
        noc_col = rank_x, noc_row = rank_y;

    nocrank = (rank_x << 8) | rank_y;
    nr_node = noc_col * (server_mode ? ++noc_row : noc_row);
    nr_node_comp = rank_x*(server_mode ? rank_y + 1 : rank_y);

    /* make nodeinfo */
    nodeinfo = new Nodeinfo[nr_node];
    for (int i = 0; i < nr_node; i++) {
        nodeinfo[i].id = entry2id(i);
        nodeinfo[i].filename = filename;
        nodeinfo[i].option = (char*)"";
        if(IDX(nodeinfo[i].id) == 0 &&
           IDY(nodeinfo[i].id) == 0)
            nodeinfo[i].type = NODE_MEM;
        else if(IDX(nodeinfo[i].id) == 0 ||
                IDY(nodeinfo[i].id) == 0)
            nodeinfo[i].type = NODE_PATH;
        else if(server_mode && IDY(nodeinfo[i].id) == rank_y + 1)
            nodeinfo[i].type = NODE_SERVER;
        else
            nodeinfo[i].type = NODE_COMP;
    }
}

/**********************************************************************/
/**
 * check arguments and set parameters
 * \param argc the number of arguments
 * \param argv arguments
 * \return 0: succeeded, 1: failed
 */
int Env::checkarg(int argc, char **argv)
{
    char *filename = NULL;
    char *b_args[MAX_NODE];
    int b_args_num = 0;
    char *O_args[MAX_NODE];
    int O_args_num = 0;

    memset(O_args, 0, MAX_NODE);
    
    for (int i = 1; argv[i] != NULL; i++) {
        char *opt = argv[i];
        if (opt[0] != '-') {
            if (filename != NULL) {
                fprintf(stderr, "## multiple binary files\n");
                return -1;
            }
            filename = opt;
            continue;
        }
        switch (opt[1]) {
        case 'b':
            b_args[b_args_num++] = opt + 2;
            break;
        case 'e':
            maxcycle = atoi_postfix(&opt[2]);
            break;
        case 'd':
            if (sscanf(opt, "-d%u", &debug_mode) != 1)
                debug_mode = 3;
            if (debug_mode > MAX_DEBUG_MODE)
                debug_mode = MAX_DEBUG_MODE;
        
            break;
        case 'q':
            quiet_mode = 1;
            break;
        case 'x':
            if (sscanf(opt, "-x%u", &rank_x) != 1)
                return -1;
            if (rank_x > MAX_COL) {
                fprintf(stderr, "## too many columns(%u)\n", 
                        rank_x);
                return -1;
            }
            break;
        case 'y':
            if (sscanf(opt, "-y%u", &rank_y) != 1)
                return -1;
            if (rank_y > MAX_ROW) {
                fprintf(stderr, "## too many rows(%u)\n", 
                        rank_y);
                return -1;
            }
            break;
        case 'A':
            server_mode = 1;
            break;
        case 'B':
            server_mode = 0;
            break;
        case 'D':
            if (sscanf(opt, "-D%u", &log_mode) != 1)
                log_mode = 1;
            if (log_mode > MAX_LOG_MODE)
                log_mode = MAX_LOG_MODE;
            {
                char *p = opt + 1;
                while(isdigit(*++p));
                if(*p != '\0')
                    logfile = p + 1;
            }
            break;
/*
        case 'm':
            multicycle = 1;
            break;
*/
        case 'O':
            O_args[O_args_num++] = opt + 2;
            break;
        case '-':
            if(!strcmp(opt + 2, "del_module")) {
                add_module = 0;
            }
            else if(!strcmp(opt + 2, "del_server")) {
                server_mode = 0;
            }
            else if(!strcmp(opt + 2, "incc_activity")) {
                incc_activity = 1;
            }
            else if(!strcmp(opt + 2, "help")) {
                return -1;
            }
            else if(!strcmp(opt + 2, "version")) {
                printf("%s %s\n", MC_NAME, MC_VER);
                exit(0);
            }
            else {
                fprintf(stderr, "## %s: not supported option!\n", opt);
                return -1;
            }
            break;
        default:
            fprintf(stderr, "## -%c: not supported option!\n", opt[1]);
            return -1;
        }
    }
    
    make_nodeinfo(filename);

    // b arg
    for (int i = 0; i < b_args_num; i++) {
        int x = 0, y = 0;
        if (sscanf(b_args[i], "%d_%d_", &x, &y) != 2 ||
            (x > rank_x || y > rank_y + add_module)) {
            fprintf(stderr, "## -b option error\n");
            fprintf(stderr, "##    id(%d,%d) rank(%d,%d)\n",
                    x, y, rank_x, rank_y);
            return -1;
        }
        char *p = b_args[i];
        while (*p++ != '_');
        while (*p++ != '_');
        nodeinfo[id2entry(x, y)].filename = p;
    }

    // o arg
    for (int i = 0; i < O_args_num; i++) {
        int x, y;
        if (sscanf(O_args[i], "%d_%d_", &x, &y) != 2 ||
            (x > noc_col - 1|| y > noc_row)) {
            fprintf(stderr, "## -o option error\n");
            fprintf(stderr, "##    id(%d,%d) rank(%d,%d)\n",
                    x, y, rank_x, rank_y);
            return -1;
        }
        char *p = O_args[i];
        while (*p++ != '_');
        while (*p++ != '_');
        nodeinfo[id2entry(x, y)].option = p;
    }

    // chk exefile
    for(int i = 0; i < nr_node; i++)
        if(!nodeinfo[i].filename)
            return -1;

    return 0;
}

/**********************************************************************/
/**
 * initialize simulation
 */
int Env::initialize(int argc, char **argv)
{
    if (checkarg(argc, argv) == -1)
        usage();
    return initialize();
}

int Env::initialize()
{
    node = new Node[nr_node];
    core = new Core[nr_node_comp];
    incc32 = new Incc32[nr_node_comp];
    incc64 = new Incc64[1];
    gmem = new GMemory[1];
    router = new Router[nr_node];
    prev_pc = new uint032_t[nr_node];

    for(int i = 0; i < nr_node; i++)
        prev_pc[i] = 0;

    for(int i = 0, c = 0, d32 = 0, d64 = 0, r = 0;
        i < nr_node; i++) {
        switch(nodeinfo[i].type) {
        case NODE_COMP:
        case NODE_SERVER:
            node[i].core = core + c++;
            {
                char buf[256];
                //sprintf(buf, "-d%u -M ../../mem_mieru.txt %s %s -f %s",
                sprintf(buf, "-d%u \"%s\" %s -f %s",
                        debug_mode, nodeinfo[i].filename,
                        nodeinfo[i].option,
                        multicycle ? "-m" : "");

                if(node[i].core->siminit(buf) != 0)
                    return -1;
            }
            if(node[i].core->mmap->next != NULL)
                node[i].mem = node[i].core->mmap->next->dev;
            else
                node[i].mem = node[i].core->mmap->dev;
            node[i].incc = incc32 + d32++;
            node[i].router = router + r++;
            break;
        case NODE_MEM:
            node[i].mem = gmem;
            node[i].incc = incc64 + d64++;
            node[i].router = router + r++;
            break;
        case NODE_PATH:
            node[i].router = router + r++;
            break;
        }
    }

    /* MClib version check */
    for(int i = 0; i < nr_node; i++) {
        if(nodeinfo[i].type != NODE_COMP &&
           nodeinfo[i].type != NODE_SERVER)
            continue;
        int check = 1;
        for(int j = 0; j < i; j++)
            if((nodeinfo[j].type == NODE_COMP ||
                nodeinfo[j].type == NODE_SERVER) &&
               !strcmp(nodeinfo[i].filename, 
                       nodeinfo[j].filename)) {
                check = 0;
                break;
            }
        if(check && check_mclib(i, nodeinfo[i].filename))
            return 1;
    }

    for (int i = 0; i < nr_node; i++) {
        if(node[i].incc != NULL)
            node[i].incc->init(this, nodeinfo[i].id);
        if(node[i].router != NULL)
            node[i].router->init(this, nodeinfo[i].id);
    }

    signal(SIGINT, sigint_handler);

    if(log_mode >= 4) {
        if(logfile == NULL && 4 == log_mode) {
            char *bin = getenv(ENV_FLIT_VIEW);
            if(bin == NULL)
                //bin = (char*)"../etc/flitview/flitview";
                bin = (char*)"flitview";
            if((fp = popen(bin, "w")) == NULL) {
                fputs("## ", stdout);
                perror(bin);
                exit(1);
            }
        }
        else {
            if((fp = fopen(logfile, "w")) == NULL)
                print_exit("file %s cannot open.\n", logfile);
        }
        fprintf(fp, "%s %s\n", MC_NAME, MC_VER);
        fprintf(fp, "%u %u\n", rank_x, rank_y + server_mode);
    }
    
    if(incc_activity) {
        if((plot = fopen("incc_activity.dat", "w")) == NULL)
            perror("cannot open filet");
        fprintf(plot, "cores:%u %u\n", noc_col, noc_row);
    }
    
    if (! quiet_mode) {
        printf("## %s %s\n", MC_NAME, MC_VER);
        printf("## [%s %s]\n", L_NAME, L_VER);
        printf("## DEBUG MODE %u, LOG MODE %u\n",
               debug_mode, log_mode);
        printf("## comp node (%d,%d) - (%d,%d)",
               1, 1, 
               rank_x, rank_y);
        if(server_mode)
            printf(" + server (%d,%d) - (%d,%d)\n",
                   1, rank_y + 1, rank_x, rank_y + 1);
        else
            putchar('\n');
        if(add_module) {
            printf("## memory node (%d,%d)\n",
                   0, 0);
            printf("## path node (%d,%d) - (%d,%d), (%d,%d) - (%d,%d)\n",
                   0, 1, 0, rank_y + server_mode,
                   1, 0, rank_x, 0);
        }
    }
    gettime();
    return 0;
}


void Env::core_update()
{
        for (int i = 0; i < nr_node; i++) {
            if (nodeinfo[i].state != 0 && node[i].core != NULL) {
                if(multicycle)
                    nodeinfo[i].state = node[i].core->chip->step_multi();
                else
                    nodeinfo[i].state = node[i].core->chip->step_funct();
            }
        }
}

void Env::memory_update()
{
        for (int i = 0; i < nr_node; i++) {
            if(node[i].core != NULL) {
                uint032_t pc;
                ((MainMemory*)(node[i].mem))->read4b(SC_MMAP_PC, &pc);
                if(pc != prev_pc[i]) { // written memory map
                    printf("### %llu memory map written as 0x%08x 0x%08x\n\r",
                           cycle,
                           pc,
                           prev_pc[i]);
//                    node[i].core->chip->mips->as->pc = pc;
//                    node[i].core->chip->mips->as->delay_npc = 0;
                }
                else
                    ((MainMemory*)node[i].mem)->
                     write4b(SC_MMAP_PC, node[i].core->chip->mips->as->pc);
                prev_pc[i] = node[i].core->chip->mips->as->pc;
            }
        }
}

void Env::incc_update()
{
        for (int i = 0; i < nr_node; i++) {
            if(node[i].incc != NULL)
                node[i].incc->update();
        }
}

void Env::router_update()
{
        for (int i = 0; i < nr_node; i++) {
            if(node[i].router != NULL)
                node[i].router->update();
        }
}

void Env::router_switching()
{
        for (int i = 0; i < nr_node; i++) {
            if(node[i].router != NULL)
                node[i].router->switching();
        }
}

void Env::comm_update()
{
        for (int i = 0; i < nr_node; i++) {
            if(node[i].incc != NULL)
                node[i].incc->comm();
            if(node[i].router != NULL)
                node[i].router->comm();
        }
}

void Env::post_update()
{
        ++cycle;

        for (int i = 0; i < nr_node; i++)
            if(node[i].core != NULL)
                ((MainMemory*)node[i].mem)->write8b(SC_MMAP_TIMER, cycle);
}



/**********************************************************************/
/**
 * main loop of the simulation
 */
void Env::execute()
{
    while (finishcond()) {
      core_update();
      memory_update();
      incc_update();
      router_update();
        print_log();
      router_switching();
        print_log();

      comm_update();

        print_log();

      post_update();
        
        fflush(stdout);
    }
}

/**********************************************************************/
/**
 * finalize simulation
 */
void Env::finalize()
{

    for (int i = 0; i < nr_node; i++)
        if(node[i].core != NULL && node[i].core->lcd != NULL)
            node[i].core->lcd->fini();


    if(log_mode > 2) {
        log_t log = {cycle, 0, 0, 0, 0, 0, 0};
        fwrite(&log, sizeof(log_t), 1, fp);
        if(logfile == NULL)
            pclose(fp);
        else
            fclose(fp);
    }

    if(incc_activity)
        fclose(plot);

    float simtime = (float)gettime() / 1000000.0f;
    if(!quiet_mode) {
        printf("## Simulation time %8.3f [sec]\n", simtime);
        printf("## Simulation cycle %8lld\n", cycle);
        printf("## Simulation %8.3f kilo cycle / sec\n",
               cycle*1e-3/simtime);
        /*
          output peak memory space used by the current process 
          if the system is unix or linux
        */
        /*
        {
            char buf[256];
            FILE *fp;
            if((fp = popen("uname", "r")) == NULL) {
                perror("uname");
                return;
            }
            fgets(buf, sizeof(buf), fp);
            pclose(fp);

            pid_t pid = getpid();
            snprintf(buf, sizeof(buf),
                     "ps vh -p %d | awk '{print $8}'", (int)pid);
            if((fp = popen(buf, "r")) == NULL) {
                perror(buf);
                return;
            }
            fgets(buf, sizeof(buf), fp);
            pclose(fp);
            float kilobyte = atoi(buf);
            char val[16];
            if((kilobyte / (1 << 20)) > 1.0)
                snprintf(val, sizeof(buf),
                         "%5.2f GB", kilobyte / (1 << 20));
            else if((kilobyte / (1 << 10)) > 1.0)
                snprintf(val, sizeof(buf),
                         "%5.2f MB", kilobyte / (1 << 10));
            else
                snprintf(val, sizeof(buf),
                         "%5.2f KB", kilobyte);
            printf("## Simulation RSS: %s\n", val);
        }
        */
    }
}

/**********************************************************************/
/**
 * return condition if simulation should be finished or not
 * \return 0: end simulation, 1: continue simulation
 */
int Env::finishcond()
{
    if(sigint) {
        puts("## interrupt");
        return 0;
    }
    if(cycle > maxcycle) {
        printf("## over maxcycle:%llu\n", maxcycle);
        return 0;
    }
    for(int i = 0; i < nr_node; i++) {
        if (nodeinfo[i].state != 0 &&
            nodeinfo[i].type == NODE_COMP &&
            ! *((char*)((Incc32*)node[i].incc)->memmap + SC_MMAP_NODE_SERVER))
            return 1;
    }
    return 0;
}

/**********************************************************************/
/**
 * convert array entry to node id
 * \param i array entry
 * \return 16bit node id
 */
uint032_t Env::entry2id(int i)
{
    uint032_t id;
    if(add_module)
        id = (((i % noc_col) & 0xff)  << 8) |
            ((i / noc_col) & 0xff);
    else
        id = ((((i % noc_col) & 0xff) + 1)  << 8) |
            (((i / noc_col) & 0xff) + 1);
    return id;
}

/**********************************************************************/
/**
 * get node id as 16 bit format
 * \param idx x-coordinate
 * \param idy y-coordinate
 * \return 16bit node id
 */
int Env::id2id(int idx, int idy)
{
    return (idx << 8) | idy;
}

/**********************************************************************/
/**
 * convert 16bit node id to array entry
 * \param id 16bit node id
 * \return array entry
 */
int Env::id2entry(uint032_t id)
{
    int entry;
    if(add_module)
        entry = noc_col * IDY(id) + IDX(id);
    else
        entry = noc_col * (IDY(id) - 1) + (IDX(id) - 1);
    return entry;
}

/**********************************************************************/
/**
 * convert x-coordinate and y-coordinate to array entry
 * \param idx x-coordinate
 * \param idy y-coordinate
 * \return array entry
 */
int Env::id2entry(int idx, int idy)
{
    int entry;
    if(add_module)
        entry = noc_col * idy + idx;
    else
        entry = noc_col * (idy - 1) + (idx - 1);
    return entry;
}

/**********************************************************************/
/**
 * get elapsed time from the last point of this function called
 * \return elapsed time (microsecond)
 */
ullint Env::gettime()
{
    static ullint start = 0;
    ullint tm;
    struct timeval tv;
    
    gettimeofday(&tv, NULL);
    tm = (ullint)tv.tv_sec * 1000000ull + tv.tv_usec;
    if (start == 0)
        start = tm;
    return tm - start;
}

/**********************************************************************/
/**
 * Nodeinfo constructor
 */
Nodeinfo::Nodeinfo()
{
    id = 0;
    type = NODE_COMP;
    state = 1;
    filename = NULL;
}

/**********************************************************************/
/**
 * Node constructor
 */
Node::Node() {
    incc = NULL;
    core = NULL;
    router = NULL;
}

/**********************************************************************/
/**
 * check MClib version
 * \param node_entry entry of node[]
 * \param filename elf file name
 * \return 0: succeeded, 1: failed
 */
int Env::check_mclib(int node_entry, const char *filename)
{
    Elf32_Shdr *str = NULL, *sym = NULL;
    char *file;
    char *progname = strtok((char*)filename, " ");

    /* open and read file */
    {
        int fd = open(progname, O_RDONLY);
        struct stat sb;
        if (fd == -1) {
            perror(progname);
            return 1;
        }
        fstat(fd, &sb);
        file = new char[sb.st_size];
        if (read(fd, file, sb.st_size) < sb.st_size) {
            perror(progname);
            return 1;
        }
        close(fd);
    }
        
    /* read ELF Header */
    {
        Elf32_Ehdr *ehdr = (Elf32_Ehdr *)file;
        
        /* read Section Header */
        for (int i = 0; i < ehdr->e_shnum; i++) {
            Elf32_Shdr *shdr = (Elf32_Shdr *)(file + ehdr->e_shoff +
                                              ehdr->e_shentsize * i);
            if (shdr->sh_type == SHT_STRTAB)
                str = shdr;
            if (shdr->sh_type == SHT_SYMTAB)
                sym = shdr;
        }
    }

    /* search symbol MC_LIB_VER */
    int ret = 0;
    for (unsigned int i = 0; i < sym->sh_size / sym->sh_entsize; i++) {
        Elf32_Sym *symp;
        symp = (Elf32_Sym *)(file + sym->sh_offset + sym->sh_entsize * i);
        if (!symp->st_name)
            continue;
        char *symname = (char *)(file + str->sh_offset + symp->st_name);
        char scan_str[] = "MC_LIB_VER";
        if(!strncmp(symname, scan_str, strlen(scan_str))) {
            uint008_t str[256];
            uint032_t ptr;
            ((MainMemory*)node[node_entry].mem)
                ->read4b((uint032_t)symp->st_value, &ptr);
            int i = 0;
            do
                ((MainMemory*)node[node_entry].mem)
                    ->read1b(ptr + i, str + i);
            while(str[i++] != '\0');
            if(strncmp(MC_LIB_VER, (char*)str + 26, strlen(MC_LIB_VER))) {
                fprintf(stderr, "%s: SimMc requires MClib version %s\n",
                       progname, MC_LIB_VER);
                ret = 1;
            }
        }
    }
    
    delete [] file;
    file = NULL;
    
    return ret;
}
