#include "network.h"

/**********************************************************************/
/**
 * GMemory constructor
 */
GMemory::GMemory() {
    // initialize list
    INIT_LIST_HEAD(&alist);
    pagebuf = NULL;
    latency_counter = 0;
}

/**********************************************************************/
/**
 * GMemory destructor
 */
GMemory::~GMemory() {
    // delete all pages
    while(!list_empty(&alist)) {
        page_t *p = list_entry(alist.next, page_t, list);
        list_del(alist.next);
        delete p;
    }
}

/**********************************************************************/
/**
 * allocate new page
 * \param addr page address
 * \return allocated page pointer
 */
GMemory::page_t *GMemory::newpage(const uint064_t addr) {
    page_t *p;
    uint064_t pagehead = addr & (~GMEM_PAGE_MASK);

    // search a page that match with page entry
    list_for_each_entry(p, &alist, list)
        if(p->pagehead == pagehead)
            return p;
    
    p = new page_t;
    p->pagehead = pagehead;
    list_add(&p->list, &alist);

    return p;
}

/**********************************************************************/
/**
 * translate an address in the simulation to the real address
 * \param addr address
 * \return real address
 */
uint032_t *GMemory::getrealaddr(const uint064_t addr) {
    uint032_t *p;
    uint032_t pageoffset = addr & GMEM_PAGE_MASK;

    p = getpagerealaddr(addr);
    
    return (uint032_t*)((char*)p + pageoffset);
}

/**********************************************************************/
/**
 * get the real address of page address
 * \param addr address
 * \return real address of the page
 */
uint032_t *GMemory::getpagerealaddr(const uint064_t addr) {
    page_t *p;
    uint064_t pagehead = addr & (~GMEM_PAGE_MASK);
    
    // search a page that match with page entry
    list_for_each_entry(p, &alist, list)
        if(p->pagehead == pagehead)
            break;

    // create page if not allocated
    if(&p->list == &alist)
        p = newpage(addr);

    return p->data;
}

/**********************************************************************/
/**
 * return whether the specified address is
 *  included in the current page buffer
 * \param addr address
 * \return 0: page fault, 1: page hit
 */
int GMemory::pagehit(const uint064_t addr) {
    uint032_t *pageaddr = getpagerealaddr(addr);
    if(pagebuf == pageaddr)
        return 1;
    if(latency_counter == PAGE_LATENCY) {
        pagebuf = pageaddr;
        latency_counter = 0;
        return 1;
    }
    latency_counter ++;
    return 0;
}

/**********************************************************************/
/**
 * read 4 byte from the specified address
 * \param addr address
 * \return data the pointer to be stored the data loaded from the memory
 */
void GMemory::read4b(const uint064_t addr, uint032_t *data) {
    *data = *getrealaddr(addr);
}

/**********************************************************************/
/**
 * write 4 byte to the specified address
 * \param addr address
 * \return data the data to be stored to the memory
 */
void GMemory::write4b(const uint064_t addr, uint032_t data) {
    *getrealaddr(addr) = data;
}

